from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import AnyStr, Dict, List, Tuple, Union
    from requests.structures import CaseInsensitiveDict
    from pyramid.config import Configurator
    from pyramid.registry import Registry
    from pyramid.request import Request as PyramidRequest
    from pyramid.response import Response as PyramidResponse
    from webob.response import Response as WebobResponse
    from webob.headers import ResponseHeaders, EnvironHeaders
    from webtest.response import TestResponse

    Number = Union[int, float]
    AnyValue = Union[AnyStr, Number, bool, None]
    AnyKey = Union[AnyStr, int]
    JSON = Dict[AnyKey, Union[AnyValue, Dict[AnyKey, 'JSON'], List['JSON']]]

    AnyContainer = Union[Configurator, Registry, PyramidRequest]
    SettingValue = Union[AnyStr, Number, bool, None]
    SettingsType = Dict[AnyStr, SettingValue]
    AnySettingsContainer = Union[AnyContainer, SettingsType]

    CookiesType = Dict[AnyStr, AnyStr]
    HeadersType = Dict[AnyStr, AnyStr]
    CookiesTupleType = List[Tuple[AnyStr, AnyStr]]
    HeadersTupleType = List[Tuple[AnyStr, AnyStr]]
    CookiesBaseType = Union[CookiesType, CookiesTupleType]
    HeadersBaseType = Union[HeadersType, HeadersTupleType]
    OptionalHeaderCookiesType = Union[Tuple[None, None], Tuple[HeadersBaseType, CookiesBaseType]]
    AnyHeadersContainer = Union[HeadersBaseType, ResponseHeaders, EnvironHeaders, CaseInsensitiveDict]
    AnyCookiesContainer = Union[CookiesBaseType, PyramidRequest, AnyHeadersContainer]
    AnyResponseType = Union[WebobResponse, PyramidResponse, TestResponse]



from twitcher.adapter.default import DefaultAdapter, TWITCHER_ADAPTER_DEFAULT
from twitcher.adapter.base import AdapterInterface
from twitcher.utils import get_settings
from importlib import import_module
from inspect import isclass
from typing import TYPE_CHECKING

import logging
LOGGER = logging.getLogger("TWITCHER")

if TYPE_CHECKING:
    from twitcher.typedefs import AnySettingsContainer
    from typing import AnyStr, Type


def import_adapter(name):
    # type: (AnyStr) -> Type[AdapterInterface]
    """Attempts import of the class specified by python string ``package.module.class``."""
    components = name.split('.')
    mod_name = components[0]
    mod = import_module(mod_name)
    for comp in components[1:]:
        if not hasattr(mod, comp):
            mod_from = mod_name
            mod_name = '{mod}.{sub}'.format(mod=mod_name, sub=comp)
            mod = import_module(mod_name, package=mod_from)
            continue
        mod = getattr(mod, comp)
        mod_name = mod.__name__
    if not isclass(mod) or not issubclass(mod, AdapterInterface):
        raise TypeError("Invalid reference is not of type '{}.{}'."
                        .format(AdapterInterface.__module__, AdapterInterface.__name__))
    return mod


def get_adapter_type(container):
    # type: (AnySettingsContainer) -> AnyStr
    """Finds the specified adapter from configuration settings."""
    settings = get_settings(container)
    return str(settings.get('twitcher.adapter', TWITCHER_ADAPTER_DEFAULT.lower()))


def get_adapter_factory(container):
    # type: (AnySettingsContainer) -> AdapterInterface
    """
    Creates an adapter interface according to `twitcher.adapter` setting.
    By default the :class:`twitcher.adapter.default.DefaultAdapter` implementation will be used.
    """
    adapter_type = get_adapter_type(container)
    if adapter_type != TWITCHER_ADAPTER_DEFAULT:
        try:
            adapter_class = import_adapter(adapter_type)
        except Exception as e:
            LOGGER.error("Adapter '{!s}' raised an exception during import : '{!r}'".format(adapter_type, e))
            raise
        try:
            LOGGER.info("Using adapter: '{!r}'".format(adapter_class))
            return adapter_class(container)
        except Exception as e:
            LOGGER.error("Adapter '{!s}' raised an exception during instantiation : '{!r}'".format(adapter_type, e))
            raise
    return DefaultAdapter(container)



"""
Read or write data from database.

Stores should not be accessed directly, but instead should use the adapter interface.

See also:
    - :class:`twitcher.adapter.base.AdapterInterface`
    - :func:`twitcher.adapter.get_adapter_factory`
"""

from twitcher.exceptions import (
    ServiceNotFound,
    DatabaseError
)
from twitcher.utils import baseurl
from twitcher import models

from sqlalchemy.exc import DBAPIError


class ServiceStore(object):
    """
    Stores a services. It inserts or updates the service with a given name.
    """
    def __init__(self, request):
        self.request = request

    def save_service(self, name, url, *args, **kwargs):
        """
        Stores an OWS service in database (insert or update).

        :param name: A service name string.
        :parm url: A URL string.
        """
        try:
            query = self.request.dbsession.query(models.Service)
            one = query.filter(models.Service.name == name).first()
            if one:
                # update
                one.url = baseurl(url)
                one.type = kwargs.get('type', 'WPS')
                one.purl = kwargs.get('purl', '')
                one._verify = int(kwargs.get('verify', 1))
                one.auth = kwargs.get('auth', 'token')
                self.request.dbsession.merge(one)
            else:
                # insert
                one = models.Service(
                    name=name,
                    url=baseurl(url),
                    type=kwargs.get('type', 'WPS'),
                    purl=kwargs.get('purl', ''),
                    _verify=int(kwargs.get('verify', 1)),
                    auth=kwargs.get('auth', 'token'))
                self.request.dbsession.add(one)
        except DBAPIError:
            raise DatabaseError
        if not one:
            raise ServiceNotFound
        return one

    def delete_service(self, name):
        """
        Removes service identified by name.
        """
        try:
            query = self.request.dbsession.query(models.Service)
            one = query.filter(models.Service.name == name).first()
        except DBAPIError:
            raise DatabaseError
        if not one:
            raise ServiceNotFound
        self.request.dbsession.delete(one)

    def list_services(self):
        """
        Lists all services.

        :return: A list with instances of :class:`twitcher.models.Service`.
        """
        try:
            services = self.request.dbsession.query(models.Service).all()
        except DBAPIError:
            raise DatabaseError
        return services

    def fetch_by_name(self, name):
        """
        Get service for given service ``name``.

        :param name: A service name string.
        :return: An instance of :class:`twitcher.models.Service`.
        """
        try:
            query = self.request.dbsession.query(models.Service)
            one = query.filter(models.Service.name == name).first()
        except DBAPIError:
            raise DatabaseError
        if not one:
            raise ServiceNotFound
        return one

    def fetch_by_url(self, url):
        """
        Get a service for given ``url``.

        :param url: A URL string.
        :return: An instance of :class:`twitcher.models.Service`.
        """
        try:
            query = self.request.dbsession.query(models.Service)
            one = query.filter(models.Service.url == url).first()
        except DBAPIError:
            raise DatabaseError
        if not one:
            raise ServiceNotFound
        return one

    def clear_services(self):
        """
        Removes all services.
        """
        try:
            self.request.dbsession.query(models.Service).delete()
        except DBAPIError:
            raise DatabaseError



# coding: utf-8
"""
twitcher.models.oauth
---------------------

The model is inspired by the flask-oauthlib example:
https://flask-oauthlib.readthedocs.io/en/latest/oauth2.html
"""

from datetime import datetime, timedelta

from sqlalchemy import (
    Column,
    Integer,
    Text,
    String,
    DateTime,
    ForeignKey
)
from sqlalchemy.orm import relationship

from .meta import Base


class Client(Base):
    __tablename__ = 'client'
    # human readable name, not required
    name = Column(String(40))
    client_id = Column(String(40), primary_key=True)
    client_secret = Column(String(55), unique=True, index=True, nullable=False)
    _redirect_uris = Column(Text)
    default_scope = Column(Text, default='compute')

    @property
    def redirect_uris(self):
        if self._redirect_uris:
            return self._redirect_uris.split()
        return []

    @property
    def default_redirect_uri(self):
        if self.redirect_uris:
            return self.redirect_uris[0]
        return ''

    @property
    def default_scopes(self):
        if self.default_scope:
            return self.default_scope.split()
        return []

    @property
    def allowed_grant_types(self):
        return ['client_credentials']


class Token(Base):
    __tablename__ = 'token'
    id = Column(Integer, primary_key=True)
    client_id = Column(
        String(40), ForeignKey('client.client_id', ondelete='CASCADE'),
        nullable=False,
    )
    client = relationship('Client')
    token_type = Column(String(40))
    access_token = Column(String(255))
    refresh_token = Column(String(255))
    expires = Column(DateTime)
    scope = Column(Text)

    def __init__(self, **kwargs):
        expires_in = kwargs.pop('expires_in', None)
        if expires_in is not None:
            self.expires = datetime.utcnow() + timedelta(seconds=expires_in)

        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def scopes(self):
        if self.scope:
            return self.scope.split()
        return []



import webtest

from .. common import BaseTest, WPS_TEST_SERVICE, dummy_request

from twitcher.store import ServiceStore


class FunctionalTest(BaseTest):
    def test_app(self):
        app = webtest.TestApp(
            self.config.make_wsgi_app(),
            extra_environ={'db.session': self.session, 'tm.active': True})
        return app

    def init_store(self):
        # add public wps service
        service_store = ServiceStore(
            dummy_request(dbsession=self.session))
        service_store.save_service(
            name="wps",
            url=WPS_TEST_SERVICE,
            type="wps",
            auth='token',
            verify=False,
            purl="http://purl/wps")
        # add secured wps service
        service_store.save_service(
            name="wps_secured",
            url=WPS_TEST_SERVICE,
            type="wps",
            auth='token',
            verify=False,
            purl="http://purl/wps_secured")



"""init

Revision ID: 0a075bcbeb8f
Revises: 1ab5b6358dd5
Create Date: 2019-06-28 17:50:57.305255

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0a075bcbeb8f'
down_revision = '1ab5b6358dd5'
branch_labels = None
depends_on = None

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('client', sa.Column('default_scope', sa.Text(), nullable=True))
    # ### end Alembic commands ###

def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('client', 'default_scope')
    # ### end Alembic commands ###



import os
import unittest
import json

from pyramid import testing

import transaction

RESOURCES_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), 'resources'))

WPS_CAPS_EMU_XML = os.path.join(RESOURCES_PATH, 'wps_caps_emu.xml')
WMS_CAPS_NCWMS2_111_XML = os.path.join(RESOURCES_PATH, 'wms_caps_ncwms2_111.xml')
WMS_CAPS_NCWMS2_130_XML = os.path.join(RESOURCES_PATH, 'wms_caps_ncwms2_130.xml')

WPS_TEST_SERVICE = os.getenv("TWITCHER_TEST_WPS", "http://localhost:5000/wps")


def dummy_request(dbsession):
    return testing.DummyRequest(dbsession=dbsession)


class BaseTest(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp(
            settings={
                'sqlalchemy.url': 'sqlite:///:memory:',
                'twitcher.username': 'testuser',
                'twitcher.password': 'testpassword',
                'twitcher.token.type': 'custom_token',
                'twitcher.token.secret': 'testsecret',
            })
        self.config.include('twitcher.models')
        settings = self.config.get_settings()

        from twitcher.models import (
            get_engine,
            get_session_factory,
            get_tm_session,
        )

        self.engine = get_engine(settings)
        session_factory = get_session_factory(self.engine)

        self.session = get_tm_session(session_factory, transaction.manager)

    def init_database(self):
        from twitcher.models.meta import Base
        Base.metadata.create_all(self.engine)
        # add test client
        from twitcher import models
        self.session.add(models.Client(client_id='dev', client_secret='dev'))

    def create_token(self):
        resp = self.app.get('/oauth/token?grant_type={}&client_id={}&client_secret={}'.format(
            'client_credentials',
            'dev',
            'dev'))
        assert resp.status_code == 200
        assert resp.content_type == 'application/json'
        resp.mustcontain('access_token')
        return json.loads(resp.text)['access_token']

    def tearDown(self):
        from twitcher.models.meta import Base

        testing.tearDown()
        transaction.abort()
        Base.metadata.drop_all(self.engine)



from twitcher import main, __version__
from twitcher.adapter.default import TWITCHER_ADAPTER_DEFAULT
from twitcher.frontpage import INFORMATION_PATH, VERSIONS_PATH

from pyramid import testing
from webtest import TestApp
import unittest


class TestFrontpageAPI(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.settings = {
            'twitcher.url': 'localhost',
            'sqlalchemy.url': 'sqlite:///:memory:',
            'twitcher.ows_security': False,
        }
        cls.config = testing.setUp(settings=cls.settings)
        cls.app = TestApp(main({}, **cls.config.registry.settings))

    def test_get_frontpage(self):
        resp = self.app.get('/')
        assert resp.status_code == 200
        assert resp.json['message'] == 'Twitcher Frontpage'
        assert resp.json['information_uri'] == self.settings['twitcher.url'] + INFORMATION_PATH
        assert resp.json['versions_uri'] == self.settings['twitcher.url'] + VERSIONS_PATH

    def test_get_info(self):
        resp = self.app.get(INFORMATION_PATH)
        assert resp.status_code == 200
        assert resp.json['author'] == __version__.__author__
        assert resp.json['doc'] == __version__.__doc__
        assert resp.json['email'] == __version__.__email__
        assert resp.json['package'] == 'twitcher'
        assert resp.json['version'] == __version__.__version__

    def test_get_versions(self):
        resp = self.app.get(VERSIONS_PATH)
        assert resp.status_code == 200
        assert isinstance(resp.json, list)
        assert resp.json[0]['name'] == 'Twitcher'
        assert resp.json[0]['type'] == 'application'
        assert resp.json[0]['version'] == __version__.__version__
        assert resp.json[1]['name'] == TWITCHER_ADAPTER_DEFAULT
        assert resp.json[1]['type'] == 'adapter'
        assert resp.json[1]['version'] == __version__.__version__



from twitcher.utils import get_settings

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from twitcher.typedefs import AnySettingsContainer, JSON
    from twitcher.interface import OWSSecurityInterface, OWSRegistryInterface
    from pyramid.config import Configurator
    from pyramid.request import Request


class AdapterInterface(object):
    """
    Common interface allowing functionality overriding using an adapter implementation.
    """
    def __init__(self, container):
        # type: (AnySettingsContainer) -> None
        self.settings = get_settings(container)

    @property
    def name(self):
        return '{}.{}'.format(self.__module__, type(self).__name__)

    def describe_adapter(self):
        # type: () -> JSON
        """
        Returns a JSON serializable dictionary describing the adapter implementation.
        """
        raise NotImplementedError

    def configurator_factory(self, container):
        # type: (AnySettingsContainer) -> Configurator
        """
        Returns the 'configurator' implementation of the adapter.
        """
        raise NotImplementedError

    def owssecurity_factory(self):
        # type: () -> OWSSecurityInterface
        """
        Returns the 'owssecurity' implementation of the adapter.
        """
        raise NotImplementedError

    def owsregistry_factory(self, request):
        # type: (Request) -> OWSRegistryInterface
        """
        Returns the 'owsregistry' implementation of the adapter.
        """
        raise NotImplementedError

    def owsproxy_config(self, container):
        # type: (AnySettingsContainer) -> None
        """
        Returns the 'owsproxy' implementation of the adapter.
        """
        raise NotImplementedError



"""
OWSExceptions are based on pyramid.httpexceptions.

See also: https://github.com/geopython/pywps/blob/master/pywps/exceptions.py
"""


from string import Template

from zope.interface import implementer

from webob import html_escape as _html_escape

from pyramid.interfaces import IExceptionResponse
from pyramid.response import Response
from pyramid.httpexceptions import (
    HTTPException,
    HTTPUnauthorized,
    HTTPBadRequest,
    HTTPInternalServerError,
    HTTPNotImplemented,
    status_map,
)

import inspect


@implementer(IExceptionResponse)
class OWSException(Response, Exception):
    status_base = HTTPNotImplemented
    code = 'NoApplicableCode'
    value = None
    locator = 'NoApplicableCode'
    explanation = 'Unknown Error'

    page_template = Template('''\
<?xml version="1.0" encoding="utf-8"?>
<ExceptionReport version="1.0.0"
    xmlns="http://www.opengis.net/ows/1.1"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xsi:schemaLocation="http://www.opengis.net/ows/1.1 http://schemas.opengis.net/ows/1.1.0/owsExceptionReport.xsd">
    <Exception exceptionCode="${code}" locator="${locator}">
        <ExceptionText>${message}</ExceptionText>
    </Exception>
</ExceptionReport>''')

    def __init__(self, detail=None, value=None, status_base=None, **kw):
        if status_base is not None:
            self.status_base = status_base
        Response.__init__(self, status=self._get_status_string(), **kw)
        Exception.__init__(self, detail)
        self.message = detail or self.explanation
        if value:
            self.locator = value

    def __str__(self, skip_body=False):
        return self.message

    def _get_status_string(self):
        if inspect.isclass(self.status_base) and issubclass(self.status_base, HTTPException):
            status = self.status_base().status
        elif isinstance(self.status_base, HTTPException):
            status = self.status_base.status
        elif isinstance(self.status_base, int):
            status = status_map[self.status_base]().status
        else:
            raise ValueError("Invalid status base configuration.")
        return status

    def prepare(self, environ):
        if not self.body:
            self.content_type = 'text/xml'
            args = {
                'code': _html_escape(self.code),
                'locator': _html_escape(self.locator),
                'message': _html_escape(self.message or ''),
            }
            page = self.page_template.substitute(args)
            page = page.encode(self.charset)
            self.app_iter = [page]
            self.body = page

    @property
    def wsgi_response(self):
        # bw compat only
        return self

    exception = wsgi_response  # bw compat only

    def __call__(self, environ, start_response):
        # differences from webob.exc.WSGIHTTPException
        #
        # - does not try to deal with HEAD requests
        #
        # - does not manufacture a new response object when generating
        #   the default response
        #
        self.prepare(environ)
        return Response.__call__(self, environ, start_response)


class OWSAccessForbidden(OWSException):
    status_base = HTTPUnauthorized
    locator = "AccessForbidden"
    explanation = "Access to this service is forbidden"


class OWSAccessFailed(OWSException):
    status_base = HTTPBadRequest
    locator = "NotAcceptable"
    explanation = "Access to this service failed"


class OWSNoApplicableCode(OWSException):
    status_base = HTTPInternalServerError


class OWSMissingParameterValue(OWSException):
    """MissingParameterValue WPS Exception"""
    status_base = HTTPBadRequest
    code = "MissingParameterValue"
    locator = ""
    explanation = "Parameter value is missing"


class OWSInvalidParameterValue(OWSException):
    """InvalidParameterValue WPS Exception"""
    status_base = HTTPBadRequest
    code = "InvalidParameterValue"
    locator = ""
    explanation = "Parameter value is invalid"



"""update service

Revision ID: d9cff565b8db
Revises: b8aa09688558
Create Date: 2019-07-24 17:10:47.137758

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd9cff565b8db'
down_revision = 'b8aa09688558'
branch_labels = None
depends_on = None

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('services', schema=None) as batch_op:
        batch_op.add_column(sa.Column('_verify', sa.Integer(), nullable=True))
        batch_op.alter_column('name',
               existing_type=sa.TEXT(),
               nullable=False)
        batch_op.alter_column('url',
               existing_type=sa.TEXT(),
               nullable=False)
        batch_op.create_index(batch_op.f('ix_services_name'), ['name'], unique=True)
        batch_op.drop_index('name_index')
        batch_op.drop_column('verify')
        batch_op.drop_column('public')

    # ### end Alembic commands ###

def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('services', schema=None) as batch_op:
        batch_op.add_column(sa.Column('public', sa.INTEGER(), nullable=True))
        batch_op.add_column(sa.Column('verify', sa.INTEGER(), nullable=True))
        batch_op.create_index('name_index', ['name'], unique=1)
        batch_op.drop_index(batch_op.f('ix_services_name'))
        batch_op.alter_column('url',
               existing_type=sa.TEXT(),
               nullable=True)
        batch_op.alter_column('name',
               existing_type=sa.TEXT(),
               nullable=True)
        batch_op.drop_column('_verify')

    # ### end Alembic commands ###



"""
The owsproxy is a service which acts as a proxy for registered OWS services.

The implementation of owsproxy is based on `papyrus_ogcproxy <https://github.com/elemoine/papyrus_ogcproxy>`_

See also: https://github.com/nive/outpost/blob/master/outpost/proxy.py
"""
import requests

from pyramid.response import Response
from pyramid.settings import asbool
from typing import TYPE_CHECKING

from twitcher.owsexceptions import OWSAccessForbidden, OWSAccessFailed
from twitcher.utils import (
    replace_caps_url,
    get_settings,
    get_twitcher_url,
    is_valid_url)

import logging
LOGGER = logging.getLogger('TWITCHER')

if TYPE_CHECKING:
    from twitcher.typedefs import AnySettingsContainer  # noqa: F401
    from pyramid.config import Configurator             # noqa: F401
    from typing import AnyStr                           # noqa: F401


allowed_content_types = (
    "application/xml",                       # XML
    "text/xml",
    "text/xml;charset=ISO-8859-1"
    "application/vnd.ogc.se_xml",            # OGC Service Exception
    "application/vnd.ogc.se+xml",            # OGC Service Exception
    # "application/vnd.ogc.success+xml",      # OGC Success (SLD Put)
    "application/vnd.ogc.wms_xml",           # WMS Capabilities
    # "application/vnd.ogc.gml",              # GML
    # "application/vnd.ogc.sld+xml",          # SLD
    "application/vnd.google-earth.kml+xml",  # KML
    "application/vnd.google-earth.kmz",
    "image/png",                             # PNG
    "image/png;mode=32bit",
    "image/gif",                             # GIF
    "image/jpeg",                            # JPEG
    "application/json",                      # JSON
    "application/json;charset=ISO-8859-1",
)

# TODO: configure allowed hosts
allowed_hosts = (
    # list allowed hosts here (no port limiting)
    # "localhost",
)


# requests.models.Response defaults its chunk size to 128 bytes, which is very slow
class BufferedResponse(object):
    def __init__(self, resp):
        self.resp = resp

    def __iter__(self):
        return self.resp.iter_content(64 * 1024)


def _send_request(request, service, extra_path=None, request_params=None):

    # TODO: fix way to build url
    url = service['url']
    if extra_path:
        url += '/' + extra_path
    if request_params:
        url += '?' + request_params
    LOGGER.debug('url = {}'.format(url))

    # forward request to target (without Host Header)
    h = dict(request.headers)
    h.pop("Host", h)
    h['Accept-Encoding'] = None
    #
    service_type = service.get('type', 'wps')
    service_verify = service.get('verify', True)
    if service_type and (service_type.lower() != 'wps'):
        try:
            resp_iter = requests.request(method=request.method.upper(), url=url, data=request.body, headers=h,
                                         stream=True, verify=service_verify)
        except Exception as e:
            return OWSAccessFailed("Request failed: {}".format(e))

        # Headers meaningful only for a single transport-level connection
        hop_by_hop = ['connection', 'keep-alive', 'public', 'proxy-authenticate', 'transfer-encoding', 'upgrade']
        return Response(app_iter=BufferedResponse(resp_iter),
                        headers={k: v for k, v in list(resp_iter.headers.items()) if k.lower() not in hop_by_hop},
                        status_code=resp_iter.status_code)
    else:
        try:
            resp = requests.request(method=request.method.upper(), url=url, data=request.body, headers=h,
                                    verify=service_verify)
        except Exception as e:
            return OWSAccessFailed("Request failed: {}".format(e))

        if resp.ok is False:
            if 'ExceptionReport' in resp.text:
                pass
            else:
                return OWSAccessFailed("Response is not ok: {}".format(resp.reason))

        # check for allowed content types
        ct = None
        # LOGGER.debug("headers=", resp.headers)
        if "Content-Type" in resp.headers:
            ct = resp.headers["Content-Type"]
            if not ct.split(";")[0] in allowed_content_types:
                msg = "Content type is not allowed: {}.".format(ct)
                LOGGER.error(msg)
                return OWSAccessForbidden(msg)
        else:
            # return OWSAccessFailed("Could not get content type from response.")
            LOGGER.warn("Could not get content type from response")

        try:
            if ct in ['text/xml', 'application/xml', 'text/xml;charset=ISO-8859-1']:
                # replace urls in xml content
                # ... if public URL is not configured use proxy url.
                if is_valid_url(service.get('purl')):
                    public_url = service['purl']
                else:
                    public_url = request.route_url('owsproxy', service_name=service['name'])
                # TODO: where do i need to replace urls?
                content = replace_caps_url(resp.content, public_url, service['url'])
            else:
                # raw content
                content = resp.content
        except Exception:
            return OWSAccessFailed("Could not decode content.")

        headers = {}
        if ct:
            headers["Content-Type"] = ct
        return Response(content, status=resp.status_code, headers=headers)


def owsproxy_base_path(container):
    # type: (AnySettingsContainer) -> AnyStr
    settings = get_settings(container)
    return settings.get('twitcher.ows_proxy_protected_path', '/ows').rstrip('/').strip()


def owsproxy_base_url(container):
    # type: (AnySettingsContainer) -> AnyStr
    twitcher_url = get_twitcher_url(container)
    owsproxy_path = owsproxy_base_path(container)
    return twitcher_url + owsproxy_path


def owsproxy_view(request):
    try:
        service_name = request.matchdict.get('service_name')
        extra_path = request.matchdict.get('extra_path')
        service = request.owsregistry.get_service_by_name(service_name)
    except Exception:
        return OWSAccessFailed("Could not find service")
    if request.is_verified is False:
        raise OWSAccessForbidden("Access to service is forbidden.")
    return _send_request(request, service, extra_path, request_params=request.query_string)


def owsproxy_defaultconfig(config):
    # type: (Configurator) -> None
    settings = get_settings(config)
    if asbool(settings.get('twitcher.ows_proxy', True)):
        protected_path = owsproxy_base_path(settings)

        config.include('twitcher.oauth2')
        config.include('twitcher.owsregistry')
        config.include('twitcher.owssecurity')
        config.add_route('owsproxy', protected_path + '/proxy/{service_name}')
        config.add_route('owsproxy_extra', protected_path + '/proxy/{service_name}/{extra_path:.*}')
        config.add_view(owsproxy_view, route_name='owsproxy')
        config.add_view(owsproxy_view, route_name='owsproxy_extra')


def includeme(config):
    from twitcher.adapter import get_adapter_factory
    get_adapter_factory(config).owsproxy_config(config)



"""
Testing the Twitcher Rest interface.
"""
import pytest

from .base import FunctionalTest


class APITest(FunctionalTest):

    def setUp(self):
        super(APITest, self).setUp()
        self.init_database()
        self.init_store()

        self.config.include('twitcher.api')
        self.app = self.test_app()

    @pytest.mark.skip(reason="not working")
    def test_register_service_and_unregister_it(self):
        # clear
        resp = self.app.get('/services/clear')
        assert resp.status_code == 200
        assert resp.content_type == 'application/json'
        assert resp is True



"""
Errors raised during the Twitcher flow.
"""


class AccessTokenNotFound(Exception):
    """
    Error indicating that an access token could not be read from the
    storage backend by an instance of :class:`twitcher.store.AccessTokenStore`.
    """
    pass


class ServiceNotFound(Exception):
    """
    Error indicating that an OWS service could not be read from the
    storage backend by an instance of :class:`twitcher.store.ServiceStore`.
    """
    pass


class ServiceRegistrationError(Exception):
    """
    Error indicating that an OWS service could not be registered in the
    storage backend by an instance of :class:`twitcher.store.ServiceStore`.
    """
    pass


class DatabaseError(Exception):
    """
    Error indicating that database could not be accesses.
    """
    pass



from twitcher.adapter import import_adapter, get_adapter_factory, TWITCHER_ADAPTER_DEFAULT
from twitcher.adapter.base import AdapterInterface
from twitcher.adapter.default import DefaultAdapter
from twitcher.interface import OWSSecurityInterface
from pyramid.testing import DummyRequest
from pathlib import Path
import pytest
import shutil
import site
import os


def test_import_adapter():
    adapter = import_adapter('twitcher.adapter.default.DefaultAdapter')
    assert adapter is DefaultAdapter, "Expect {!s}, but got {!s}".format(DefaultAdapter, adapter)
    assert isinstance(adapter({}), AdapterInterface), \
        "Expect {!s}, but got {!s}".format(AdapterInterface, type(adapter))


def test_adapter_factory_default_explicit():
    settings = {'twitcher.adapter': TWITCHER_ADAPTER_DEFAULT}
    adapter = get_adapter_factory(settings)
    assert isinstance(adapter, DefaultAdapter), "Expect {!s}, but got {!s}".format(DefaultAdapter, type(adapter))


def test_adapter_factory_none_specified():
    adapter = get_adapter_factory({})
    assert isinstance(adapter, DefaultAdapter), "Expect {!s}, but got {!s}".format(DefaultAdapter, type(adapter))


# noinspection PyAbstractClass,PyMethodMayBeStatic
class DummyAdapter(AdapterInterface):
    def owssecurity_factory(self, request):
        class DummyOWSSecurity(OWSSecurityInterface):
            def verify_request(self, request): return True   # noqa: E704
        return DummyOWSSecurity()


# noinspection PyPep8Naming
def test_adapter_factory_DummyAdapter_valid_import_with_init():
    settings = {'twitcher.adapter': DummyAdapter({}).name}
    adapter = get_adapter_factory(settings)
    assert isinstance(adapter, DummyAdapter), "Expect {!s}, but got {!s}".format(DummyAdapter, type(adapter))


def make_path(base, other='__init__.py'):
    return str(Path(base) / Path(other))


def test_adapter_factory_TmpAdapter_valid_import_installed_without_init():
    """
    Test a valid installed adapter import by setting although not located under same directory,
    with much deeper structure, and without any '__init__.py' defining a python 'package'.
    """

    mod_pkg = 'test_package'
    mod_base = site.getsitepackages()[0]
    try:
        mod_name = '{}.module.submodule.section'.format(mod_pkg)
        mod_path = make_path(mod_base, mod_name.replace('.', '/'))
        mod_file_name = 'file'
        mod_file = make_path(mod_path, '{}.py'.format(mod_file_name))
        mod_class = 'TmpAdapter'
        os.makedirs(mod_path, exist_ok=True)
        with open(mod_file, 'w') as f:
            f.writelines([
                "from twitcher.adapter.base import AdapterInterface\n\n",
                "class {}(AdapterInterface):\n".format(mod_class),
                "    pass\n"
            ])
        mod_import = '.'.join([mod_name, mod_file_name, mod_class])

        settings = {'twitcher.adapter': mod_import}
        adapter = get_adapter_factory(settings)
        adapter_mod_name = '.'.join([adapter.__module__, type(adapter).__name__])
        assert not isinstance(adapter, DefaultAdapter)
        assert isinstance(adapter, AdapterInterface)
        assert adapter_mod_name == mod_import
    finally:
        shutil.rmtree(make_path(mod_base, mod_pkg), ignore_errors=True)


# noinspection PyAbstractClass
class DummyAdapterFake(object):
    pass


# noinspection PyPep8Naming
def test_adapter_factory_TestAdapter_invalid_raised():
    # fake adapter doesn't have 'name' property, provide at least same functionality
    dummy_name = '{}.{}'.format(DummyAdapterFake.__module__, DummyAdapterFake.__name__)
    settings = {'twitcher.adapter': dummy_name}
    with pytest.raises(TypeError) as err:
        get_adapter_factory(settings)
        pytest.fail(msg="Invalid adapter not inheriting from 'AdapterInterface' should raise on import.")
    adapter_str = '{}.{}'.format(AdapterInterface.__module__, AdapterInterface.__name__)
    assert adapter_str in str(err.value), "Expected to have full adapter import string in error message."


# noinspection PyTypeChecker
def test_adapter_factory_call_owssecurity_factory():
    settings = {'twitcher.adapter': DummyAdapter({}).name}
    adapter = get_adapter_factory(settings)
    security = adapter.owssecurity_factory(DummyRequest())
    assert isinstance(security, OWSSecurityInterface)
    assert security.verify_request(DummyRequest()) is True, "Requested adapter should have been called."



"""
Run tests OWSProxy tests with external WPS.

Please start `Emu WPS <https://emu.readthedocs.io/en/latest/>`_ on port 5000:

http://localhost:5000/wps
"""

import pytest

from .base import FunctionalTest


class OWSProxyAppTest(FunctionalTest):

    def setUp(self):
        super(OWSProxyAppTest, self).setUp()
        self.init_database()
        self.init_store()

        self.config.include('twitcher.owsproxy')
        self.app = self.test_app()

    @pytest.mark.online
    def test_getcaps(self):
        resp = self.app.get('/ows/proxy/wps?service=wps&request=getcapabilities')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        resp.mustcontain('</wps:Capabilities>')

    @pytest.mark.online
    def test_describeprocess(self):
        resp = self.app.get(
            '/ows/proxy/wps?service=wps&request=describeprocess&version=1.0.0&identifier=dummyprocess')
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        resp.mustcontain('</wps:ProcessDescriptions>')

    @pytest.mark.online
    def test_execute_allowed(self):
        access_token = self.create_token()
        url = "/ows/proxy/wps_secured?service=wps&request=execute&version=1.0.0&identifier=hello&DataInputs=name=tux&access_token={}".format(access_token)  # noqa
        resp = self.app.get(url)
        assert resp.status_code == 200
        assert resp.content_type == 'text/xml'
        print(resp.body)
        resp.mustcontain(
            '<wps:ProcessSucceeded>PyWPS Process Say Hello finished</wps:ProcessSucceeded>')



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()

about = {}
with open(os.path.join(here, 'twitcher', '__version__.py'), 'r') as f:
    exec(f.read(), about)

reqs = [line.strip() for line in open('requirements.txt')]
dev_reqs = [line.strip() for line in open('requirements_dev.txt')]

setup(name='pyramid_twitcher',
      version=about['__version__'],
      description=about['__doc__'],
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
          "Programming Language :: Python",
          "Framework :: Pyramid",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
          "Development Status :: 4 - Beta",
      ],
      author=about['__author__'],
      author_email=about['__email__'],
      url='https://github.com/bird-house/twitcher.git',
      license='Apache License 2.0',
      keywords='web pyramid twitcher birdhouse wps security proxy ows ogc',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='twitcher',
      install_requires=reqs,
      extras_require={
          "dev": dev_reqs,              # pip install ".[dev]"
          "postgres": ["psycopg2"],     # when using postgres database driver with sqlalchemy
      },
      entry_points="""\
      [paste.app_factory]
      main = twitcher:main
      [console_scripts]
      twitcherctl=twitcher.scripts.twitcherctl:main
      initialize_twitcher_db=twitcher.scripts.initialize_db:main
      """,
      )



from pyramid.settings import asbool

from twitcher.interface import OWSSecurityInterface
from twitcher.owsrequest import OWSRequest
from twitcher.utils import get_settings


class OWSSecurity(OWSSecurityInterface):
    def verify_request(self, request):
        """Verify that the service request is allowed.

        This method verifies that the provided credentials are valid.
        Depending on the authentication configuration this could be
        a client X509 certificate or an OAuth2 token.
        """
        ows_request = OWSRequest(request)
        if ows_request.service_allowed() is False:
            return False
        try:
            service_name = request.matchdict.get('service_name')
            service = request.owsregistry.get_service_by_name(service_name)
        except Exception:
            return False
        if service.get('public', False) is True:
            return True
        if ows_request.public_access() is True:
            return True
        if service.get('auth', '') == 'cert':
            # Check the verification result of the client certificate.
            # Verifcation is done by nginx.
            return request.headers.get('X-Ssl-Client-Verify', '') == 'SUCCESS'
        else:
            # verify the oauth token for compute scope.
            return request.verify_request(scopes=["compute"])


def includeme(config):
    from twitcher.adapter import get_adapter_factory
    settings = get_settings(config)
    security_enabled = asbool(settings.get('twitcher.ows_security', True))

    def is_verified(request):
        if not security_enabled:
            return True
        adapter = get_adapter_factory(request)
        return adapter.owssecurity_factory().verify_request(request)
    config.add_request_method(is_verified, reify=True)



from pyramid.config import Configurator
from pyramid.request import Request
from pyramid.registry import Registry
from datetime import datetime
from urllib import parse as urlparse
from lxml import etree
from typing import TYPE_CHECKING
import json
import time
import pytz
import re

from twitcher.exceptions import ServiceNotFound

import logging
LOGGER = logging.getLogger("TWITCHER")

if TYPE_CHECKING:
    from twitcher.typedefs import AnySettingsContainer, SettingsType
    from typing import AnyStr, Optional


def get_settings(container):
    # type: (AnySettingsContainer) -> Optional[SettingsType]
    """
    Retrieves the application ``settings`` from various containers referencing to it.

    :raises TypeError: if the container type cannot be identified to retrieve settings.
    """
    if isinstance(container, (Configurator, Request)):
        container = container.registry
    if isinstance(container, Registry):
        container = container.settings
    if isinstance(container, dict):
        return container
    raise TypeError("Could not retrieve settings from container object of type [{}]".format(type(container)))


def get_twitcher_url(container):
    # type: (AnySettingsContainer) -> AnyStr
    settings = get_settings(container)
    return settings.get('twitcher.url').rstrip('/').strip()


def sanitize(name, minlen=2, maxlen=25):
    """Lower-case name and replace all non-ascii chars by `_`."""
    if name is None or len(name.strip()) < minlen:
        raise ValueError("name must have at least {} chars.".format(minlen))
    return re.sub(r'\W|^(?=\d)', '_', name.strip().lower()[:maxlen])


def is_valid_url(url):
    try:
        parsed_url = urlparse.urlparse(url)
        return True if all([parsed_url.scheme, ]) else False
    except Exception:
        return False


def is_json_serializable(item):
    try:
        json.dumps(item)
        return True
    except (TypeError, OverflowError):
        return False


def parse_service_name(url, protected_path):
    parsed_url = urlparse.urlparse(url)
    service_name = None
    if parsed_url.path.startswith(protected_path):
        parts_without_protected_path = parsed_url.path[len(protected_path)::].strip('/').split('/')
        if 'proxy' in parts_without_protected_path:
            parts_without_protected_path.remove('proxy')
        if len(parts_without_protected_path) > 0:
            service_name = parts_without_protected_path[0]
    if not service_name:
        raise ServiceNotFound
    return service_name


def now():
    return localize_datetime(datetime.utcnow())


def now_secs():
    """
    Return the current time in seconds since the Epoch.
    """
    return int(time.time())


def expires_at(hours=1):
    return now_secs() + hours * 3600


def localize_datetime(dt, tz_name='UTC'):
    """Provide a timezone-aware object for a given datetime and timezone name
    """
    tz_aware_dt = dt
    if dt.tzinfo is None:
        utc = pytz.timezone('UTC')
        aware = utc.localize(dt)
        timezone = pytz.timezone(tz_name)
        tz_aware_dt = aware.astimezone(timezone)
    else:
        LOGGER.warning('tzinfo already set')
    return tz_aware_dt


def baseurl(url):
    """
    return baseurl of given url
    """
    parsed_url = urlparse.urlparse(url)
    if not parsed_url.netloc or parsed_url.scheme not in ("http", "https"):
        raise ValueError('bad url')
    service_url = "%s://%s%s" % (parsed_url.scheme, parsed_url.netloc, parsed_url.path.strip())
    return service_url


def path_elements(path):
    elements = [el.strip() for el in path.split('/')]
    elements = [el for el in elements if len(el) > 0]
    return elements


def lxml_strip_ns(tree):
    for node in tree.iter():
        try:
            has_namespace = node.tag.startswith('{')
        except AttributeError:
            continue  # node.tag is not a string (node is a comment or similar)
        if has_namespace:
            node.tag = node.tag.split('}', 1)[1]


def replace_caps_url(xml, url, prev_url=None):
    ns = {
        'ows': 'http://www.opengis.net/ows/1.1',
        'xlink': 'http://www.w3.org/1999/xlink'}
    doc = etree.fromstring(xml)
    # wms 1.1.1 onlineResource
    if 'WMT_MS_Capabilities' in doc.tag:
        LOGGER.debug("replace proxy urls in wms 1.1.1")
        for element in doc.findall('.//OnlineResource[@xlink:href]', namespaces=ns):
            parsed_url = urlparse.urlparse(element.get('{http://www.w3.org/1999/xlink}href'))
            new_url = url
            if parsed_url.query:
                new_url += '?' + parsed_url.query
            element.set('{http://www.w3.org/1999/xlink}href', new_url)
        xml = etree.tostring(doc)
    # wms 1.3.0 onlineResource
    elif 'WMS_Capabilities' in doc.tag:
        LOGGER.debug("replace proxy urls in wms 1.3.0")
        for element in doc.findall('.//{http://www.opengis.net/wms}OnlineResource[@xlink:href]', namespaces=ns):
            parsed_url = urlparse.urlparse(element.get('{http://www.w3.org/1999/xlink}href'))
            new_url = url
            if parsed_url.query:
                new_url += '?' + parsed_url.query
            element.set('{http://www.w3.org/1999/xlink}href', new_url)
        xml = etree.tostring(doc)
    # wps operations
    elif 'Capabilities' in doc.tag:
        for element in doc.findall('ows:OperationsMetadata//*[@xlink:href]', namespaces=ns):
            element.set('{http://www.w3.org/1999/xlink}href', url)
        xml = etree.tostring(doc)
    elif prev_url:
        xml = xml.decode('utf-8', 'ignore')
        xml = xml.replace(prev_url, url)
    return xml



"""
Basic Authentication.

Taken from:
https://docs.pylonsproject.org/projects/pyramid-cookbook/en/latest/auth/basic.html
"""
from pyramid.settings import asbool
from pyramid.authentication import BasicAuthAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.httpexceptions import HTTPForbidden
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.security import (
    ALL_PERMISSIONS,
    Allow,
    Authenticated,
    forget)
from pyramid.view import forbidden_view_config

from twitcher.utils import get_settings


@forbidden_view_config()
def forbidden_view(request):
    if request.authenticated_userid is None:
        response = HTTPUnauthorized()
        response.headers.update(forget(request))

    # user is logged in but doesn't have permissions, reject wholesale
    else:
        response = HTTPForbidden()
    return response


def check_credentials(username, password, request):
    settings = request.registry.settings
    _username = settings['twitcher.username']
    _password = settings['twitcher.password']
    if username == _username and password == _password:
        # an empty list is enough to indicate logged-in
        return []


class Root:
    # dead simple, give everyone who is logged in any permission
    __acl__ = (
        (Allow, Authenticated, ALL_PERMISSIONS),
    )


def includeme(config):
    settings = get_settings(config)
    if asbool(settings.get('twitcher.basicauth', True)):
        authn_policy = BasicAuthAuthenticationPolicy(check=check_credentials, debug=True)
        authz_policy = ACLAuthorizationPolicy()
        config.set_authorization_policy(authz_policy)
        config.set_authentication_policy(authn_policy)
        config.set_root_factory(lambda request: Root())
